@@ -28,10 +28,10 @@
 #include <stdlib.h>
 #include <glib/gi18n-lib.h>
 
+#include "cogl.h"
+#include "cogl-private.h"
 #include "cogl-debug.h"
 
-#ifdef COGL_ENABLE_DEBUG
-
 /* XXX: If you add a debug option, please also add an option
  * definition to cogl-debug-options.h. This will enable us - for
  * example - to emit a "help" description for the option.
@@ -83,7 +83,7 @@ static const GDebugKey cogl_behavioural_debug_keys[] = {
 static const int n_cogl_behavioural_debug_keys =
   G_N_ELEMENTS (cogl_behavioural_debug_keys);
 
-unsigned int _cogl_debug_flags[COGL_DEBUG_N_INTS];
+unsigned long _cogl_debug_flags[COGL_DEBUG_N_LONGS];
 GHashTable *_cogl_debug_instances;
 
 static void
@@ -92,41 +92,59 @@ _cogl_parse_debug_string_for_keys (const char *value,
                                    const GDebugKey *keys,
                                    unsigned int nkeys)
 {
-  int int_num, key_num;
+  int long_num, key_num;
 
   /* g_parse_debug_string expects the value field in GDebugKey to be a
-     mask in a guint but we may have multiple guints so we need to
-     build a separate array for each possible guint */
+     mask in a guint but the flags is stored in an array of multiple
+     longs so we need to build a separate array for each possible
+     guint */
 
-  for (int_num = 0; int_num < COGL_DEBUG_N_INTS; int_num++)
+  for (long_num = 0; long_num < COGL_DEBUG_N_LONGS; long_num++)
     {
-      GDebugKey keys_for_int[sizeof (unsigned int) * 8];
-      unsigned int mask_for_int;
-      int nkeys_for_int = 0;
+      int int_num;
 
-      for (key_num = 0; key_num < nkeys; key_num++)
-        if (COGL_FLAGS_GET_INDEX (keys[key_num].value) == int_num)
-          {
-            keys_for_int[nkeys_for_int] = keys[key_num];
-            keys_for_int[nkeys_for_int].value =
-              COGL_FLAGS_GET_MASK (keys[key_num].value);
-            nkeys_for_int++;
-          }
-
-      if (nkeys_for_int > 0)
+      for (int_num = 0;
+           int_num < sizeof (unsigned long) / sizeof (unsigned int);
+           int_num++)
         {
-          mask_for_int = g_parse_debug_string (value,
-                                               keys_for_int,
-                                               nkeys_for_int);
-          if (enable)
-            _cogl_debug_flags[int_num] |= mask_for_int;
-          else
-            _cogl_debug_flags[int_num] &= ~mask_for_int;
+          GDebugKey keys_for_int[sizeof (unsigned int) * 8];
+          int nkeys_for_int = 0;
+
+          for (key_num = 0; key_num < nkeys; key_num++)
+            {
+              int long_index = COGL_FLAGS_GET_INDEX (keys[key_num].value);
+              int int_index = (keys[key_num].value %
+                               (sizeof (unsigned long) * 8) /
+                               (sizeof (unsigned int) * 8));
+
+              if (long_index == long_num && int_index == int_num)
+                {
+                  keys_for_int[nkeys_for_int] = keys[key_num];
+                  keys_for_int[nkeys_for_int].value =
+                    COGL_FLAGS_GET_MASK (keys[key_num].value) >>
+                    (int_num * sizeof (unsigned int) * 8);
+                  nkeys_for_int++;
+                }
+            }
+
+          if (nkeys_for_int > 0)
+            {
+              unsigned long mask =
+                ((unsigned long) g_parse_debug_string (value,
+                                                       keys_for_int,
+                                                       nkeys_for_int)) <<
+                (int_num * sizeof (unsigned int) * 8);
+
+              if (enable)
+                _cogl_debug_flags[long_num] |= mask;
+              else
+                _cogl_debug_flags[long_num] &= ~mask;
+            }
         }
     }
 }
 
-static void
+void
 _cogl_parse_debug_string (const char *value,
                           gboolean enable,
                           gboolean ignore_help)
@@ -149,17 +167,18 @@ _cogl_parse_debug_string (const char *value,
         else
           COGL_DEBUG_CLEAR_FLAG (cogl_log_debug_keys[i].value);
     }
-  else if (strcmp (value, "help") == 0)
+  else if (g_ascii_strcasecmp (value, "help") == 0)
     {
-      g_printerr ("\n\n%28s\n", "Supported debug values:");
+      g_printerr ("\n\n%28s\n", _("Supported debug values:"));
 #define OPT(MASK_NAME, GROUP, NAME, NAME_FORMATTED, DESCRIPTION) \
-      g_printerr ("%28s %s\n", NAME ":", DESCRIPTION);
+      g_printerr ("%28s %s\n", NAME ":", g_dgettext (GETTEXT_PACKAGE, \
+                                                     DESCRIPTION));
 #include "cogl-debug-options.h"
-      g_printerr ("\n%28s\n", "Special debug values:");
+      g_printerr ("\n%28s\n", _("Special debug values:"));
       OPT (IGNORED, "ignored", "all", "ignored", \
-           "Enables all non-behavioural debug options");
+           N_("Enables all non-behavioural debug options"));
       OPT (IGNORED, "ignored", "verbose", "ignored", \
-           "Enables all non-behavioural debug options");
+           N_("Enables all non-behavioural debug options"));
 #undef OPT
       exit (1);
     }
@@ -176,6 +195,7 @@ _cogl_parse_debug_string (const char *value,
     }
 }
 
+#ifdef COGL_ENABLE_DEBUG
 static gboolean
 cogl_arg_debug_cb (const char *key,
                    const char *value,
@@ -209,13 +229,9 @@ static GOptionEntry cogl_args[] = {
   { NULL, },
 };
 
-static gboolean
-pre_parse_hook (GOptionContext  *context,
-                GOptionGroup    *group,
-                gpointer         data,
-                GError         **error)
+void
+_cogl_debug_check_environment (void)
 {
-#ifdef COGL_ENABLE_DEBUG
   const char *env_string;
 
   env_string = g_getenv ("COGL_DEBUG");
@@ -226,11 +242,31 @@ pre_parse_hook (GOptionContext  *context,
                                 FALSE /* don't ignore help */);
       env_string = NULL;
     }
-#endif /* COGL_ENABLE_DEBUG */
+
+  env_string = g_getenv ("COGL_NO_DEBUG");
+  if (env_string != NULL)
+    {
+      _cogl_parse_debug_string (env_string,
+                                FALSE /* disable the flags */,
+                                FALSE /* don't ignore help */);
+      env_string = NULL;
+    }
+}
+
+static gboolean
+pre_parse_hook (GOptionContext  *context,
+                GOptionGroup    *group,
+                gpointer         data,
+                GError         **error)
+{
+  _cogl_init ();
 
   return TRUE;
 }
 
+/* XXX: GOption based library initialization is not reliable because the
+ * GOption API has no way to represent dependencies between libraries.
+ */
 GOptionGroup *
 cogl_get_option_group (void)
 {
