@@ -292,8 +292,7 @@ _cogl_journal_flush_modelview_and_entries (CoglJournalEntry *batch_start,
     {
       _cogl_matrix_stack_set (state->modelview_stack,
                               &batch_start->model_view);
-      _cogl_matrix_stack_flush_to_gl (state->modelview_stack,
-                                      COGL_MATRIX_MODELVIEW);
+      _cogl_context_set_current_modelview (ctx, state->modelview_stack);
     }
 
   attributes = (CoglAttribute **)state->attributes->data;
@@ -709,6 +708,12 @@ _cogl_journal_flush_clip_stacks_and_entries (CoglJournalEntry *batch_start,
 
   _cogl_clip_stack_flush (batch_start->clip_stack, state->framebuffer);
 
+  /* XXX: Because we are manually flushing clip state here we need to
+   * make sure that the clip state gets updated the next time we flush
+   * framebuffer state by marking the current framebuffer's clip state
+   * as changed. */
+  ctx->current_draw_buffer_changes |= COGL_FRAMEBUFFER_STATE_CLIP;
+
   _cogl_matrix_stack_push (state->modelview_stack);
 
   /* If we have transformed all our quads at log time then we ensure
@@ -718,15 +723,13 @@ _cogl_journal_flush_clip_stacks_and_entries (CoglJournalEntry *batch_start,
   if (G_LIKELY (!(COGL_DEBUG_ENABLED (COGL_DEBUG_DISABLE_SOFTWARE_TRANSFORM))))
     {
       _cogl_matrix_stack_load_identity (state->modelview_stack);
-      _cogl_matrix_stack_flush_to_gl (state->modelview_stack,
-                                      COGL_MATRIX_MODELVIEW);
+      _cogl_context_set_current_modelview (ctx, state->modelview_stack);
     }
 
   /* Setting up the clip state can sometimes also flush the projection
      matrix so we should flush it again. This will be a no-op if the
      clip code didn't modify the projection */
-  _cogl_matrix_stack_flush_to_gl (state->projection_stack,
-                                  COGL_MATRIX_PROJECTION);
+  _cogl_context_set_current_projection (ctx, state->projection_stack);
 
   batch_and_call (batch_start,
                   batch_len,
@@ -1370,8 +1373,9 @@ _cogl_journal_flush (CoglJournal *journal,
      state manually */
   _cogl_framebuffer_flush_state (framebuffer,
                                  framebuffer,
-                                 COGL_FRAMEBUFFER_FLUSH_SKIP_MODELVIEW |
-                                 COGL_FRAMEBUFFER_FLUSH_SKIP_CLIP_STATE);
+                                 COGL_FRAMEBUFFER_STATE_ALL &
+                                 ~(COGL_FRAMEBUFFER_STATE_MODELVIEW |
+                                   COGL_FRAMEBUFFER_STATE_CLIP));
 
   state.journal = journal;
 
