@@ -58,10 +58,6 @@ struct _CoglMatrixStack
 
   GArray *stack;
 
-  /* which state does GL have, NULL if unknown */
-  CoglMatrixState *flushed_state;
-  gboolean flushed_identity;
-
   unsigned int age;
 };
 
@@ -124,9 +120,6 @@ _cogl_matrix_stack_top_mutable (CoglMatrixStack *stack,
         cogl_matrix_init_identity (&new_top->matrix);
       else
         new_top->matrix = state->matrix;
-
-      if (stack->flushed_state == state)
-        stack->flushed_state = new_top;
     }
 
   return new_top;
@@ -191,11 +184,6 @@ _cogl_matrix_stack_pop (CoglMatrixStack *stack)
           return;
         }
 
-      if (stack->flushed_state == state)
-        {
-          stack->flushed_state = NULL;
-        }
-
       stack->age++;
       g_array_set_size (stack->stack, stack->stack->len - 1);
     }
@@ -218,9 +206,6 @@ _cogl_matrix_stack_load_identity (CoglMatrixStack *stack)
   if (!state->is_identity)
     {
       state->is_identity = TRUE;
-
-      /* mark dirty */
-      stack->flushed_state = NULL;
       stack->age++;
     }
 }
@@ -235,8 +220,6 @@ _cogl_matrix_stack_scale (CoglMatrixStack *stack,
 
   state = _cogl_matrix_stack_top_mutable (stack, TRUE);
   cogl_matrix_scale (&state->matrix, x, y, z);
-  /* mark dirty */
-  stack->flushed_state = NULL;
   state->is_identity = FALSE;
   stack->age++;
 }
@@ -251,8 +234,6 @@ _cogl_matrix_stack_translate (CoglMatrixStack *stack,
 
   state = _cogl_matrix_stack_top_mutable (stack, TRUE);
   cogl_matrix_translate (&state->matrix, x, y, z);
-  /* mark dirty */
-  stack->flushed_state = NULL;
   state->is_identity = FALSE;
   stack->age++;
 }
@@ -268,8 +249,6 @@ _cogl_matrix_stack_rotate (CoglMatrixStack *stack,
 
   state = _cogl_matrix_stack_top_mutable (stack, TRUE);
   cogl_matrix_rotate (&state->matrix, angle, x, y, z);
-  /* mark dirty */
-  stack->flushed_state = NULL;
   state->is_identity = FALSE;
   stack->age++;
 }
@@ -282,8 +261,6 @@ _cogl_matrix_stack_multiply (CoglMatrixStack  *stack,
 
   state = _cogl_matrix_stack_top_mutable (stack, TRUE);
   cogl_matrix_multiply (&state->matrix, &state->matrix, matrix);
-  /* mark dirty */
-  stack->flushed_state = NULL;
   state->is_identity = FALSE;
   stack->age++;
 }
@@ -303,8 +280,6 @@ _cogl_matrix_stack_frustum (CoglMatrixStack *stack,
   cogl_matrix_frustum (&state->matrix,
                        left, right, bottom, top,
                        z_near, z_far);
-  /* mark dirty */
-  stack->flushed_state = NULL;
   state->is_identity = FALSE;
   stack->age++;
 }
@@ -321,8 +296,6 @@ _cogl_matrix_stack_perspective (CoglMatrixStack *stack,
   state = _cogl_matrix_stack_top_mutable (stack, TRUE);
   cogl_matrix_perspective (&state->matrix,
                            fov_y, aspect, z_near, z_far);
-  /* mark dirty */
-  stack->flushed_state = NULL;
   state->is_identity = FALSE;
   stack->age++;
 }
@@ -341,8 +314,6 @@ _cogl_matrix_stack_ortho (CoglMatrixStack *stack,
   state = _cogl_matrix_stack_top_mutable (stack, TRUE);
   cogl_matrix_ortho (&state->matrix,
                      left, right, bottom, top, z_near, z_far);
-  /* mark dirty */
-  stack->flushed_state = NULL;
   state->is_identity = FALSE;
   stack->age++;
 }
@@ -387,158 +358,122 @@ _cogl_matrix_stack_set (CoglMatrixStack  *stack,
 
   state = _cogl_matrix_stack_top_mutable (stack, FALSE);
   state->matrix = *matrix;
-  /* mark dirty */
-  stack->flushed_state = NULL;
   state->is_identity = FALSE;
   stack->age++;
 }
 
-#if defined (HAVE_COGL_GL) || defined (HAVE_COGL_GLES)
-
 static void
-flush_to_fixed_api_gl (gboolean is_identity,
-                       const CoglMatrix *matrix,
-                       void *user_data)
+_cogl_matrix_stack_flush_matrix_to_gl_builtin (CoglContext *ctx,
+                                               gboolean is_identity,
+                                               CoglMatrix *matrix,
+                                               CoglMatrixMode mode)
 {
-  CoglMatrixStack *stack = user_data;
+  g_assert (ctx->driver == COGL_DRIVER_GL ||
+            ctx->driver == COGL_DRIVER_GLES1);
 
-  _COGL_GET_CONTEXT (ctx, NO_RETVAL);
-
-  if (is_identity)
+#if defined (HAVE_COGL_GL) || defined (HAVE_COGL_GLES)
+  if (ctx->flushed_matrix_mode != mode)
     {
-      if (!stack->flushed_identity)
-        GE (ctx, glLoadIdentity ());
-      stack->flushed_identity = TRUE;
-    }
-  else
-    {
-      GE (ctx, glLoadMatrixf (cogl_matrix_get_array (matrix)) );
-      stack->flushed_identity = FALSE;
-    }
-}
-
-#endif
-
-void
-_cogl_matrix_stack_prepare_for_flush (CoglMatrixStack *stack,
-                                      CoglMatrixMode mode,
-                                      CoglMatrixStackFlushFunc callback,
-                                      void *user_data)
-{
-  CoglMatrixState *state;
-
-  _COGL_GET_CONTEXT (ctx, NO_RETVAL);
-
-  state = _cogl_matrix_stack_top (stack);
-
-  /* Because Cogl defines texture coordinates to have a top left origin and
-   * because offscreen framebuffers may be used for rendering to textures we
-   * always render upside down to offscreen buffers.
-   */
-  if (mode == COGL_MATRIX_PROJECTION &&
-      cogl_is_offscreen (cogl_get_draw_framebuffer ()))
-    {
-      CoglMatrix flipped_projection;
-      CoglMatrix *projection =
-        state->is_identity ? &ctx->identity_matrix : &state->matrix;
-
-      cogl_matrix_multiply (&flipped_projection,
-                            &ctx->y_flip_matrix, projection);
-      callback (FALSE, &flipped_projection, user_data);
-    }
-  else
-    callback (state->is_identity,
-              state->is_identity ? &ctx->identity_matrix : &state->matrix,
-              user_data);
-}
-
-void
-_cogl_matrix_stack_flush_to_gl (CoglMatrixStack *stack,
-                                CoglMatrixMode   mode)
-{
-  CoglMatrixState *state;
-
-  _COGL_GET_CONTEXT (ctx, NO_RETVAL);
-
-  state = _cogl_matrix_stack_top (stack);
-
-  if (ctx->driver == COGL_DRIVER_GLES2)
-    {
-      /* Under GLES2 we need to flush the matrices differently because
-         they are stored in uniforms attached to the program instead of
-         the global GL context state. At this point we can't be sure that
-         the right program will be generated so instead we'll just store a
-         reference to the matrix stack that is intended to be flushed and
-         update the uniform once the program is ready. */
+      GLenum gl_mode = 0;
 
       switch (mode)
         {
         case COGL_MATRIX_MODELVIEW:
-          cogl_object_ref (stack);
-          if (ctx->flushed_modelview_stack)
-            cogl_object_unref (ctx->flushed_modelview_stack);
-          ctx->flushed_modelview_stack = stack;
+          gl_mode = GL_MODELVIEW;
           break;
 
         case COGL_MATRIX_PROJECTION:
-          cogl_object_ref (stack);
-          if (ctx->flushed_projection_stack)
-            cogl_object_unref (ctx->flushed_projection_stack);
-          ctx->flushed_projection_stack = stack;
+          gl_mode = GL_PROJECTION;
           break;
 
         case COGL_MATRIX_TEXTURE:
-          /* This shouldn't happen because the texture matrices are
-             handled by the GLSL pipeline backend */
-          g_assert_not_reached ();
+          gl_mode = GL_TEXTURE;
           break;
         }
+
+      GE (ctx, glMatrixMode (gl_mode));
+      ctx->flushed_matrix_mode = mode;
     }
 
-#if defined (HAVE_COGL_GL) || defined (HAVE_COGL_GLES)
+  if (is_identity)
+    GE (ctx, glLoadIdentity ());
   else
-    {
-      if (stack->flushed_state == state)
-        return;
-
-      if (ctx->flushed_matrix_mode != mode)
-        {
-          GLenum gl_mode = 0;
-
-          switch (mode)
-            {
-            case COGL_MATRIX_MODELVIEW:
-              gl_mode = GL_MODELVIEW;
-              break;
-
-            case COGL_MATRIX_PROJECTION:
-              gl_mode = GL_PROJECTION;
-              break;
-
-            case COGL_MATRIX_TEXTURE:
-              gl_mode = GL_TEXTURE;
-              break;
-            }
-
-          GE (ctx, glMatrixMode (gl_mode));
-          ctx->flushed_matrix_mode = mode;
-        }
-
-      _cogl_matrix_stack_prepare_for_flush (stack,
-                                            mode,
-                                            flush_to_fixed_api_gl,
-                                            stack);
-    }
+    GE (ctx, glLoadMatrixf (cogl_matrix_get_array (matrix)));
 #endif
-
-  stack->flushed_state = state;
 }
 
 void
-_cogl_matrix_stack_dirty (CoglMatrixStack *stack)
+_cogl_matrix_stack_flush_to_gl_builtins (CoglContext *ctx,
+                                         CoglMatrixStack *stack,
+                                         CoglMatrixMode mode,
+                                         gboolean disable_flip)
 {
-  stack->flushed_state = NULL;
-  stack->flushed_identity = FALSE;
+  g_assert (ctx->driver == COGL_DRIVER_GL ||
+            ctx->driver == COGL_DRIVER_GLES1);
+
+#if defined (HAVE_COGL_GL) || defined (HAVE_COGL_GLES)
+  {
+    gboolean needs_flip;
+    CoglMatrixState *state;
+    CoglMatrixStackCache *cache;
+
+    state = _cogl_matrix_stack_top (stack);
+
+    if (mode == COGL_MATRIX_PROJECTION)
+      {
+        /* Because Cogl defines texture coordinates to have a top left
+         * origin and because offscreen framebuffers may be used for
+         * rendering to textures we always render upside down to
+         * offscreen buffers. Also for some backends we need to render
+         * onscreen buffers upside-down too.
+         */
+        if (disable_flip)
+          needs_flip = FALSE;
+        else
+          needs_flip = cogl_is_offscreen (cogl_get_draw_framebuffer ());
+
+        cache = &ctx->builtin_flushed_projection;
+      }
+    else
+      {
+        needs_flip = FALSE;
+
+        if (mode == COGL_MATRIX_MODELVIEW)
+          cache = &ctx->builtin_flushed_modelview;
+        else
+          cache = NULL;
+      }
+
+    /* We don't need to do anything if the state is the same */
+    if (!cache ||
+        _cogl_matrix_stack_check_and_update_cache (stack, cache, needs_flip))
+      {
+        gboolean is_identity = state->is_identity && !needs_flip;
+
+        if (needs_flip)
+          {
+            CoglMatrix flipped_matrix;
+
+            cogl_matrix_multiply (&flipped_matrix,
+                                  &ctx->y_flip_matrix,
+                                  state->is_identity ?
+                                  &ctx->identity_matrix :
+                                  &state->matrix);
+
+            _cogl_matrix_stack_flush_matrix_to_gl_builtin (ctx,
+                                                           /* not identity */
+                                                           FALSE,
+                                                           &flipped_matrix,
+                                                           mode);
+          }
+        else
+          _cogl_matrix_stack_flush_matrix_to_gl_builtin (ctx,
+                                                         is_identity,
+                                                         &state->matrix,
+                                                         mode);
+      }
+  }
+#endif
 }
 
 unsigned int
@@ -552,3 +487,67 @@ _cogl_matrix_stack_has_identity_flag (CoglMatrixStack *stack)
 {
   return _cogl_matrix_stack_top (stack)->is_identity;
 }
+
+gboolean
+_cogl_matrix_stack_equal (CoglMatrixStack *stack0,
+                          CoglMatrixStack *stack1)
+{
+  CoglMatrixState *state0 = _cogl_matrix_stack_top (stack0);
+  CoglMatrixState *state1 = _cogl_matrix_stack_top (stack1);
+
+  if (state0->is_identity != state1->is_identity)
+    return FALSE;
+
+  if (state0->is_identity)
+    return TRUE;
+  else
+    return cogl_matrix_equal (&state0->matrix, &state1->matrix);
+}
+
+gboolean
+_cogl_matrix_stack_check_and_update_cache (CoglMatrixStack *stack,
+                                           CoglMatrixStackCache *cache,
+                                           gboolean flip)
+{
+  gboolean is_identity =
+    _cogl_matrix_stack_has_identity_flag (stack) && !flip;
+  gboolean is_dirty;
+
+  if (is_identity && cache->flushed_identity)
+    is_dirty = FALSE;
+  else if (cache->stack == NULL ||
+           cache->stack->age != cache->age ||
+           flip != cache->flipped)
+    is_dirty = TRUE;
+  else
+    is_dirty = (cache->stack != stack &&
+                !_cogl_matrix_stack_equal (cache->stack, stack));
+
+  /* We'll update the cache values even if the stack isn't dirty in
+     case the reason it wasn't dirty is because we compared the
+     matrices and found them to be the same. In that case updating the
+     cache values will avoid the comparison next time */
+  cache->age = stack->age;
+  cogl_object_ref (stack);
+  if (cache->stack)
+    cogl_object_unref (cache->stack);
+  cache->stack = stack;
+  cache->flushed_identity = is_identity;
+  cache->flipped = flip;
+
+  return is_dirty;
+}
+
+void
+_cogl_matrix_stack_init_cache (CoglMatrixStackCache *cache)
+{
+  cache->stack = NULL;
+  cache->flushed_identity = FALSE;
+}
+
+void
+_cogl_matrix_stack_destroy_cache (CoglMatrixStackCache *cache)
+{
+  if (cache->stack)
+    cogl_object_unref (cache->stack);
+}
