@@ -86,3 +86,11 @@ cogl_onscreen_template_set_samples_per_pixel (
 {
   onscreen_template->config.samples_per_pixel = samples_per_pixel;
 }
+
+void
+cogl_onscreen_template_set_swap_throttled (
+                                          CoglOnscreenTemplate *onscreen_template,
+                                          gboolean throttled)
+{
+  onscreen_template->config.swap_throttled = throttled;
+}
