@@ -37,6 +37,7 @@
 #include "cogl-pipeline-layer-private.h"
 #include "cogl-shader-private.h"
 #include "cogl-blend-string.h"
+#include "cogl-snippet-private.h"
 
 #ifdef COGL_PIPELINE_FRAGEND_GLSL
 
@@ -181,6 +182,38 @@ _cogl_pipeline_fragend_glsl_get_shader (CoglPipeline *pipeline)
     return 0;
 }
 
+static CoglPipelineSnippetList *
+get_fragment_snippets (CoglPipeline *pipeline)
+{
+  pipeline =
+    _cogl_pipeline_get_authority (pipeline,
+                                  COGL_PIPELINE_STATE_FRAGMENT_SNIPPETS);
+
+  return &pipeline->big_state->fragment_snippets;
+}
+
+static CoglPipelineSnippetList *
+get_layer_fragment_snippets (CoglPipelineLayer *layer)
+{
+  unsigned long state = COGL_PIPELINE_LAYER_STATE_FRAGMENT_SNIPPETS;
+  layer = _cogl_pipeline_layer_get_authority (layer, state);
+
+  return &layer->big_state->fragment_snippets;
+}
+
+static gboolean
+has_replace_hook (CoglPipelineLayer *layer,
+                  CoglSnippetHook hook)
+{
+  CoglPipelineSnippet *snippet;
+
+  COGL_LIST_FOREACH (snippet, get_layer_fragment_snippets (layer), list_node)
+    if (snippet->snippet->hook == hook && snippet->snippet->replace)
+      return TRUE;
+
+  return FALSE;
+}
+
 static gboolean
 _cogl_pipeline_fragend_glsl_start (CoglPipeline *pipeline,
                                    int n_layers,
@@ -316,7 +349,7 @@ _cogl_pipeline_fragend_glsl_start (CoglPipeline *pipeline,
 
   g_string_append (shader_state->source,
                    "void\n"
-                   "main ()\n"
+                   "cogl_generated_source ()\n"
                    "{\n");
 
   for (i = 0; i < n_layers; i++)
@@ -336,16 +369,7 @@ add_constant_lookup (CoglPipelineShaderState *shader_state,
 {
   int unit_index = _cogl_pipeline_layer_get_unit_index (layer);
 
-  /* Create a sampler uniform for this layer if we haven't already */
-  if (!shader_state->unit_state[unit_index].combine_constant_used)
-    {
-      g_string_append_printf (shader_state->header,
-                              "uniform vec4 _cogl_layer_constant_%i;\n",
-                              unit_index);
-      shader_state->unit_state[unit_index].combine_constant_used = TRUE;
-    }
-
-  g_string_append_printf (shader_state->source,
+  g_string_append_printf (shader_state->header,
                           "_cogl_layer_constant_%i.%s",
                           unit_index, swizzle);
 }
@@ -355,9 +379,8 @@ ensure_texture_lookup_generated (CoglPipelineShaderState *shader_state,
                                  CoglPipeline *pipeline,
                                  CoglPipelineLayer *layer)
 {
-  CoglHandle texture;
   int unit_index = _cogl_pipeline_layer_get_unit_index (layer);
-  const char *target_string, *tex_coord_swizzle;
+  CoglPipelineSnippetData snippet_data;
 
   _COGL_GET_CONTEXT (ctx, NO_RETVAL);
 
@@ -366,70 +389,14 @@ ensure_texture_lookup_generated (CoglPipelineShaderState *shader_state,
 
   shader_state->unit_state[unit_index].sampled = TRUE;
 
-  g_string_append_printf (shader_state->source,
-                          "  vec4 texel%i = ",
-                          unit_index);
-
-  if (G_UNLIKELY (COGL_DEBUG_ENABLED (COGL_DEBUG_DISABLE_TEXTURING)))
-    {
-      g_string_append (shader_state->source,
-                       "vec4 (1.0, 1.0, 1.0, 1.0);\n");
-
-      return;
-    }
-
-  texture = _cogl_pipeline_layer_get_texture (layer);
-
-  if (texture == NULL)
-    {
-      target_string = "2D";
-      tex_coord_swizzle = "st";
-    }
-  else
-    {
-      GLenum gl_target;
-
-      cogl_texture_get_gl_texture (texture, NULL, &gl_target);
-      switch (gl_target)
-        {
-#ifdef HAVE_COGL_GL
-        case GL_TEXTURE_1D:
-          target_string = "1D";
-          tex_coord_swizzle = "s";
-          break;
-#endif
-
-        case GL_TEXTURE_2D:
-          target_string = "2D";
-          tex_coord_swizzle = "st";
-          break;
-
-#ifdef GL_ARB_texture_rectangle
-        case GL_TEXTURE_RECTANGLE_ARB:
-          target_string = "2DRect";
-          tex_coord_swizzle = "st";
-          break;
-#endif
-
-        case GL_TEXTURE_3D:
-          target_string = "3D";
-          tex_coord_swizzle = "stp";
-          break;
-
-        default:
-          g_assert_not_reached ();
-        }
-    }
-
-  /* Create a sampler uniform */
   g_string_append_printf (shader_state->header,
-                          "uniform sampler%s _cogl_sampler_%i;\n",
-                          target_string,
+                          "vec4 cogl_texel%i;\n",
                           unit_index);
 
   g_string_append_printf (shader_state->source,
-                          "texture%s (_cogl_sampler_%i, ",
-                          target_string, unit_index);
+                          "  cogl_texel%i = cogl_texture_lookup%i (",
+                          unit_index,
+                          unit_index);
 
   /* If point sprite coord generation is being used then divert to the
      built-in varying var for that instead of the texture
@@ -443,14 +410,108 @@ ensure_texture_lookup_generated (CoglPipelineShaderState *shader_state,
       cogl_pipeline_get_layer_point_sprite_coords_enabled (pipeline,
                                                            layer->index))
     g_string_append_printf (shader_state->source,
-                            "gl_PointCoord.%s",
-                            tex_coord_swizzle);
+                            "gl_PointCoord");
   else
     g_string_append_printf (shader_state->source,
-                            "cogl_tex_coord_in[%d].%s",
-                            unit_index, tex_coord_swizzle);
+                            "cogl_tex_coord_in[%d]",
+                            unit_index);
 
   g_string_append (shader_state->source, ");\n");
+
+  /* There's no need to generate the real texture lookup if it's going
+     to be replaced */
+  if (!has_replace_hook (layer, COGL_SNIPPET_HOOK_TEXTURE_LOOKUP))
+    {
+      CoglHandle texture = _cogl_pipeline_layer_get_texture (layer);
+      const char *target_string, *tex_coord_swizzle;
+
+      if (texture == NULL)
+        {
+          target_string = "2D";
+          tex_coord_swizzle = "st";
+        }
+      else
+        {
+          GLenum gl_target;
+
+          cogl_texture_get_gl_texture (texture, NULL, &gl_target);
+          switch (gl_target)
+            {
+#ifdef HAVE_COGL_GL
+            case GL_TEXTURE_1D:
+              target_string = "1D";
+              tex_coord_swizzle = "s";
+              break;
+#endif
+
+            case GL_TEXTURE_2D:
+              target_string = "2D";
+              tex_coord_swizzle = "st";
+              break;
+
+#ifdef GL_ARB_texture_rectangle
+            case GL_TEXTURE_RECTANGLE_ARB:
+              target_string = "2DRect";
+              tex_coord_swizzle = "st";
+              break;
+#endif
+
+            case GL_TEXTURE_3D:
+              target_string = "3D";
+              tex_coord_swizzle = "stp";
+              break;
+
+            default:
+              g_assert_not_reached ();
+            }
+        }
+
+      /* Create a sampler uniform */
+      if (G_LIKELY (!COGL_DEBUG_ENABLED (COGL_DEBUG_DISABLE_TEXTURING)))
+        g_string_append_printf (shader_state->header,
+                                "uniform sampler%s _cogl_sampler_%i;\n",
+                                target_string,
+                                unit_index);
+
+      g_string_append_printf (shader_state->header,
+                              "vec4\n"
+                              "cogl_real_texture_lookup%i (vec4 coords)\n"
+                              "{\n"
+                              "  return ",
+                              unit_index);
+
+      if (G_UNLIKELY (COGL_DEBUG_ENABLED (COGL_DEBUG_DISABLE_TEXTURING)))
+        g_string_append (shader_state->header,
+                         "vec4 (1.0, 1.0, 1.0, 1.0);\n");
+      else
+        g_string_append_printf (shader_state->header,
+                                "texture%s (_cogl_sampler_%i, coords.%s);\n",
+                                target_string, unit_index, tex_coord_swizzle);
+
+      g_string_append (shader_state->header, "}\n");
+    }
+
+  /* Wrap the texture lookup in any snippets that have been hooked */
+  memset (&snippet_data, 0, sizeof (snippet_data));
+  snippet_data.snippets = get_layer_fragment_snippets (layer);
+  snippet_data.hook = COGL_SNIPPET_HOOK_TEXTURE_LOOKUP;
+  snippet_data.chain_function = g_strdup_printf ("cogl_real_texture_lookup%i",
+                                                 unit_index);
+  snippet_data.final_name = g_strdup_printf ("cogl_texture_lookup%i",
+                                             unit_index);
+  snippet_data.function_prefix = g_strdup_printf ("cogl_texture_lookup_hook%i",
+                                                  unit_index);
+  snippet_data.return_type = "vec4";
+  snippet_data.return_variable = "cogl_texel";
+  snippet_data.arguments = "cogl_tex_coord";
+  snippet_data.argument_declarations = "vec4 cogl_tex_coord";
+  snippet_data.source_buf = shader_state->header;
+
+  _cogl_pipeline_snippet_generate_code (&snippet_data);
+
+  g_free ((char *) snippet_data.chain_function);
+  g_free ((char *) snippet_data.final_name);
+  g_free ((char *) snippet_data.function_prefix);
 }
 
 static void
@@ -462,7 +523,7 @@ add_arg (CoglPipelineShaderState *shader_state,
          CoglPipelineCombineOp operand,
          const char *swizzle)
 {
-  GString *shader_source = shader_state->source;
+  GString *shader_source = shader_state->header;
   char alpha_swizzle[5] = "aaaa";
 
   g_string_append_c (shader_source, '(');
@@ -486,7 +547,7 @@ add_arg (CoglPipelineShaderState *shader_state,
     {
     case COGL_PIPELINE_COMBINE_SOURCE_TEXTURE:
       g_string_append_printf (shader_source,
-                              "texel%i.%s",
+                              "cogl_texel%i.%s",
                               _cogl_pipeline_layer_get_unit_index (layer),
                               swizzle);
       break;
@@ -502,7 +563,7 @@ add_arg (CoglPipelineShaderState *shader_state,
       if (previous_layer_index >= 0)
         {
           g_string_append_printf (shader_source,
-                                  "layer%i.%s",
+                                  "cogl_layer%i.%s",
                                   previous_layer_index,
                                   swizzle);
           break;
@@ -516,7 +577,7 @@ add_arg (CoglPipelineShaderState *shader_state,
       if (src >= COGL_PIPELINE_COMBINE_SOURCE_TEXTURE0 &&
           src < COGL_PIPELINE_COMBINE_SOURCE_TEXTURE0 + 32)
         g_string_append_printf (shader_source,
-                                "texel%i.%s",
+                                "cogl_texel%i.%s",
                                 src - COGL_PIPELINE_COMBINE_SOURCE_TEXTURE0,
                                 swizzle);
       break;
@@ -560,9 +621,22 @@ ensure_arg_generated (CoglPipeline *pipeline,
 
   switch (src)
     {
-    case COGL_PIPELINE_COMBINE_SOURCE_CONSTANT:
     case COGL_PIPELINE_COMBINE_SOURCE_PRIMARY_COLOR:
-      /* These don't involve any other layers */
+      /* This doesn't involve any other layers */
+      break;
+
+    case COGL_PIPELINE_COMBINE_SOURCE_CONSTANT:
+      {
+        int unit_index = _cogl_pipeline_layer_get_unit_index (layer);
+        /* Create a sampler uniform for this layer if we haven't already */
+        if (!shader_state->unit_state[unit_index].combine_constant_used)
+          {
+            g_string_append_printf (shader_state->header,
+                                    "uniform vec4 _cogl_layer_constant_%i;\n",
+                                    unit_index);
+            shader_state->unit_state[unit_index].combine_constant_used = TRUE;
+          }
+      }
       break;
 
     case COGL_PIPELINE_COMBINE_SOURCE_PREVIOUS:
@@ -597,6 +671,20 @@ ensure_arg_generated (CoglPipeline *pipeline,
     }
 }
 
+static void
+ensure_args_for_func (CoglPipeline *pipeline,
+                      CoglPipelineLayer *layer,
+                      int previous_layer_index,
+                      CoglPipelineCombineFunc function,
+                      CoglPipelineCombineSource *src)
+{
+  int n_args = _cogl_get_n_args_for_combine_func (function);
+  int i;
+
+  for (i = 0; i < n_args; i++)
+    ensure_arg_generated (pipeline, layer, previous_layer_index, src[i]);
+}
+
 static void
 append_masked_combine (CoglPipeline *pipeline,
                        CoglPipelineLayer *layer,
@@ -607,18 +695,10 @@ append_masked_combine (CoglPipeline *pipeline,
                        CoglPipelineCombineOp *op)
 {
   CoglPipelineShaderState *shader_state = get_shader_state (pipeline);
-  GString *shader_source = shader_state->source;
-  int n_args;
-  int i;
+  GString *shader_source = shader_state->header;
 
-  n_args = _cogl_get_n_args_for_combine_func (function);
-
-  for (i = 0; i < n_args; i++)
-    ensure_arg_generated (pipeline, layer, previous_layer_index, src[i]);
-
-  g_string_append_printf (shader_state->source,
-                          "  layer%i.%s = ",
-                          layer->index,
+  g_string_append_printf (shader_state->header,
+                          "  cogl_layer.%s = ",
                           swizzle);
 
   switch (function)
@@ -715,6 +795,7 @@ ensure_layer_generated (CoglPipeline *pipeline,
   CoglPipelineLayer *combine_authority;
   CoglPipelineLayerBigState *big_state;
   CoglPipelineLayer *layer;
+  CoglPipelineSnippetData snippet_data;
   LayerData *layer_data;
 
   /* Find the layer that corresponds to this layer_num */
@@ -740,41 +821,98 @@ ensure_layer_generated (CoglPipeline *pipeline,
                                         COGL_PIPELINE_LAYER_STATE_COMBINE);
   big_state = combine_authority->big_state;
 
-  g_string_append_printf (shader_state->source,
-                          "  vec4 layer%i;\n",
+  /* Make a global variable for the result of the layer code */
+  g_string_append_printf (shader_state->header,
+                          "vec4 cogl_layer%i;\n",
                           layer_index);
 
-  if (!_cogl_pipeline_layer_needs_combine_separate (combine_authority) ||
-      /* GL_DOT3_RGBA Is a bit weird as a GL_COMBINE_RGB function
-       * since if you use it, it overrides your ALPHA function...
-       */
-      big_state->texture_combine_rgb_func ==
-      COGL_PIPELINE_COMBINE_FUNC_DOT3_RGBA)
-    append_masked_combine (pipeline,
-                           layer,
-                           layer_data->previous_layer_index,
-                           "rgba",
-                           big_state->texture_combine_rgb_func,
-                           big_state->texture_combine_rgb_src,
-                           big_state->texture_combine_rgb_op);
-  else
+  /* Skip the layer generation if there is a snippet that replaces the
+     default layer code. This is important because generating this
+     code may cause the code for other layers to be generated and
+     stored in the global variable. If this code isn't actually used
+     then the global variables would be uninitialised and they may be
+     used from other layers */
+  if (!has_replace_hook (layer, COGL_SNIPPET_HOOK_LAYER_FRAGMENT))
     {
-      append_masked_combine (pipeline,
-                             layer,
-                             layer_data->previous_layer_index,
-                             "rgb",
-                             big_state->texture_combine_rgb_func,
-                             big_state->texture_combine_rgb_src,
-                             big_state->texture_combine_rgb_op);
-      append_masked_combine (pipeline,
-                             layer,
-                             layer_data->previous_layer_index,
-                             "a",
-                             big_state->texture_combine_alpha_func,
-                             big_state->texture_combine_alpha_src,
-                             big_state->texture_combine_alpha_op);
+      ensure_args_for_func (pipeline,
+                            layer,
+                            layer_data->previous_layer_index,
+                            big_state->texture_combine_rgb_func,
+                            big_state->texture_combine_rgb_src);
+      ensure_args_for_func (pipeline,
+                            layer,
+                            layer_data->previous_layer_index,
+                            big_state->texture_combine_alpha_func,
+                            big_state->texture_combine_alpha_src);
+
+      g_string_append_printf (shader_state->header,
+                              "vec4\n"
+                              "cogl_real_generate_layer%i ()\n"
+                              "{\n"
+                              "  vec4 cogl_layer;\n",
+                              layer_index);
+
+      if (!_cogl_pipeline_layer_needs_combine_separate (combine_authority) ||
+          /* GL_DOT3_RGBA Is a bit weird as a GL_COMBINE_RGB function
+           * since if you use it, it overrides your ALPHA function...
+           */
+          big_state->texture_combine_rgb_func ==
+          COGL_PIPELINE_COMBINE_FUNC_DOT3_RGBA)
+        append_masked_combine (pipeline,
+                               layer,
+                               layer_data->previous_layer_index,
+                               "rgba",
+                               big_state->texture_combine_rgb_func,
+                               big_state->texture_combine_rgb_src,
+                               big_state->texture_combine_rgb_op);
+      else
+        {
+          append_masked_combine (pipeline,
+                                 layer,
+                                 layer_data->previous_layer_index,
+                                 "rgb",
+                                 big_state->texture_combine_rgb_func,
+                                 big_state->texture_combine_rgb_src,
+                                 big_state->texture_combine_rgb_op);
+          append_masked_combine (pipeline,
+                                 layer,
+                                 layer_data->previous_layer_index,
+                                 "a",
+                                 big_state->texture_combine_alpha_func,
+                                 big_state->texture_combine_alpha_src,
+                                 big_state->texture_combine_alpha_op);
+        }
+
+      g_string_append (shader_state->header,
+                       "  return cogl_layer;\n"
+                       "}\n");
     }
 
+  /* Wrap the layer code in any snippets that have been hooked */
+  memset (&snippet_data, 0, sizeof (snippet_data));
+  snippet_data.snippets = get_layer_fragment_snippets (layer);
+  snippet_data.hook = COGL_SNIPPET_HOOK_LAYER_FRAGMENT;
+  snippet_data.chain_function = g_strdup_printf ("cogl_real_generate_layer%i",
+                                                 layer_index);
+  snippet_data.final_name = g_strdup_printf ("cogl_generate_layer%i",
+                                             layer_index);
+  snippet_data.function_prefix = g_strdup_printf ("cogl_generate_layer%i",
+                                                  layer_index);
+  snippet_data.return_type = "vec4";
+  snippet_data.return_variable = "cogl_layer";
+  snippet_data.source_buf = shader_state->header;
+
+  _cogl_pipeline_snippet_generate_code (&snippet_data);
+
+  g_free ((char *) snippet_data.chain_function);
+  g_free ((char *) snippet_data.final_name);
+  g_free ((char *) snippet_data.function_prefix);
+
+  g_string_append_printf (shader_state->source,
+                          "  cogl_layer%i = cogl_generate_layer%i ();\n",
+                          layer_index,
+                          layer_index);
+
   g_slice_free (LayerData, layer_data);
 }
 
@@ -885,6 +1023,7 @@ _cogl_pipeline_fragend_glsl_end (CoglPipeline *pipeline,
       GLint lengths[2];
       GLint compile_status;
       GLuint shader;
+      CoglPipelineSnippetData snippet_data;
 
       COGL_STATIC_COUNTER (fragend_glsl_compile_counter,
                            "glsl fragment compile counter",
@@ -906,7 +1045,7 @@ _cogl_pipeline_fragend_glsl_end (CoglPipeline *pipeline,
 
           ensure_layer_generated (pipeline, last_layer->index);
           g_string_append_printf (shader_state->source,
-                                  "  cogl_color_out = layer%i;\n",
+                                  "  cogl_color_out = cogl_layer%i;\n",
                                   last_layer->index);
 
           COGL_LIST_FOREACH_SAFE (layer_data, &shader_state->layers,
@@ -922,8 +1061,19 @@ _cogl_pipeline_fragend_glsl_end (CoglPipeline *pipeline,
         add_alpha_test_snippet (pipeline, shader_state);
 #endif
 
+      /* Close the function surrounding the generated fragment processing */
       g_string_append (shader_state->source, "}\n");
 
+      /* Add all of the hooks for fragment processing */
+      memset (&snippet_data, 0, sizeof (snippet_data));
+      snippet_data.snippets = get_fragment_snippets (pipeline);
+      snippet_data.hook = COGL_SNIPPET_HOOK_FRAGMENT;
+      snippet_data.chain_function = "cogl_generated_source";
+      snippet_data.final_name = "main";
+      snippet_data.function_prefix = "cogl_fragment_hook";
+      snippet_data.source_buf = shader_state->source;
+      _cogl_pipeline_snippet_generate_code (&snippet_data);
+
       GE_RET( shader, ctx, glCreateShader (GL_FRAGMENT_SHADER) );
 
       lengths[0] = shader_state->header->len;
