@@ -117,7 +117,6 @@ typedef enum {
 } CoglPipelineWrapMode;
 /* NB: these values come from the equivalents in gl.h */
 
-#define cogl_pipeline_set_layer_texture cogl_pipeline_set_layer_texture_EXP
 /**
  * cogl_pipeline_set_layer:
  * @pipeline: A #CoglPipeline object
@@ -136,25 +135,26 @@ typedef enum {
  * as purely GLSL based layers.</note>
  *
  * Since: 2.0
+ * Stability: unstable
  */
 void
 cogl_pipeline_set_layer_texture (CoglPipeline *pipeline,
                                  int           layer_index,
                                  CoglTexture  *texture);
 
-#define cogl_pipeline_remove_layer cogl_pipeline_remove_layer_EXP
 /**
  * cogl_pipeline_remove_layer:
  * @pipeline: A #CoglPipeline object
  * @layer_index: Specifies the layer you want to remove
  *
  * This function removes a layer from your pipeline
+ * Since: 1.10
+ * Stability: unstable
  */
 void
 cogl_pipeline_remove_layer (CoglPipeline *pipeline,
 			    int           layer_index);
 
-#define cogl_pipeline_set_layer_combine cogl_pipeline_set_layer_combine_EXP
 /**
  * cogl_pipeline_set_layer_combine:
  * @pipeline: A #CoglPipeline object
@@ -246,6 +246,7 @@ cogl_pipeline_remove_layer (CoglPipeline *pipeline,
  *   or hardware. On failure, %FALSE is returned and @error is set
  *
  * Since: 2.0
+ * Stability: unstable
  */
 gboolean
 cogl_pipeline_set_layer_combine (CoglPipeline *pipeline,
@@ -253,8 +254,6 @@ cogl_pipeline_set_layer_combine (CoglPipeline *pipeline,
 				 const char   *blend_string,
                                  GError      **error);
 
-#define cogl_pipeline_set_layer_combine_constant \
-  cogl_pipeline_set_layer_combine_constant_EXP
 /**
  * cogl_pipeline_set_layer_combine_constant:
  * @pipeline: A #CoglPipeline object
@@ -266,13 +265,13 @@ cogl_pipeline_set_layer_combine (CoglPipeline *pipeline,
  * description then you can use this function to define its value.
  *
  * Since: 2.0
+ * Stability: unstable
  */
 void
 cogl_pipeline_set_layer_combine_constant (CoglPipeline    *pipeline,
                                           int              layer_index,
                                           const CoglColor *constant);
 
-#define cogl_pipeline_set_layer_matrix cogl_pipeline_set_layer_matrix_EXP
 /**
  * cogl_pipeline_set_layer_matrix:
  * @pipeline: A #CoglPipeline object
@@ -281,13 +280,15 @@ cogl_pipeline_set_layer_combine_constant (CoglPipeline    *pipeline,
  *
  * This function lets you set a matrix that can be used to e.g. translate
  * and rotate a single layer of a pipeline used to fill your geometry.
+ *
+ * Since: 1.10
+ * Stability: unstable
  */
 void
 cogl_pipeline_set_layer_matrix (CoglPipeline     *pipeline,
 				int               layer_index,
 				const CoglMatrix *matrix);
 
-#define cogl_pipeline_get_n_layers cogl_pipeline_get_n_layers_EXP
 /**
  * cogl_pipeline_get_n_layers:
  * @pipeline: A #CoglPipeline object
@@ -297,11 +298,11 @@ cogl_pipeline_set_layer_matrix (CoglPipeline     *pipeline,
  * Return value: the number of layers
  *
  * Since: 2.0
+ * Stability: unstable
  */
 int
 cogl_pipeline_get_n_layers (CoglPipeline *pipeline);
 
-#define cogl_pipeline_set_layer_filters cogl_pipeline_set_layer_filters_EXP
 /**
  * cogl_pipeline_set_layer_filters:
  * @pipeline: A #CoglPipeline object
@@ -311,6 +312,9 @@ cogl_pipeline_get_n_layers (CoglPipeline *pipeline);
  *
  * Changes the decimation and interpolation filters used when a texture is
  * drawn at other scales than 100%.
+ *
+ * Since: 1.10
+ * Stability: unstable
  */
 void
 cogl_pipeline_set_layer_filters (CoglPipeline      *pipeline,
@@ -318,8 +322,6 @@ cogl_pipeline_set_layer_filters (CoglPipeline      *pipeline,
                                  CoglPipelineFilter min_filter,
                                  CoglPipelineFilter mag_filter);
 
-#define cogl_pipeline_set_layer_point_sprite_coords_enabled \
-  cogl_pipeline_set_layer_point_sprite_coords_enabled_EXP
 /**
  * cogl_pipeline_set_layer_point_sprite_coords_enabled:
  * @pipeline: a #CoglHandle to a pipeline.
@@ -340,6 +342,7 @@ cogl_pipeline_set_layer_filters (CoglPipeline      *pipeline,
  *
  * Return value: %TRUE if the function succeeds, %FALSE otherwise.
  * Since: 2.0
+ * Stability: unstable
  */
 gboolean
 cogl_pipeline_set_layer_point_sprite_coords_enabled (CoglPipeline *pipeline,
@@ -347,8 +350,6 @@ cogl_pipeline_set_layer_point_sprite_coords_enabled (CoglPipeline *pipeline,
                                                      gboolean      enable,
                                                      GError      **error);
 
-#define cogl_pipeline_get_layer_point_sprite_coords_enabled \
-  cogl_pipeline_get_layer_point_sprite_coords_enabled_EXP
 /**
  * cogl_pipeline_get_layer_point_sprite_coords_enabled:
  * @pipeline: a #CoglHandle to a pipeline.
@@ -361,13 +362,12 @@ cogl_pipeline_set_layer_point_sprite_coords_enabled (CoglPipeline *pipeline,
  * point sprite coordinates.
  *
  * Since: 2.0
+ * Stability: unstable
  */
 gboolean
 cogl_pipeline_get_layer_point_sprite_coords_enabled (CoglPipeline *pipeline,
                                                      int           layer_index);
 
-#define cogl_pipeline_get_layer_wrap_mode_s \
-  cogl_pipeline_get_layer_wrap_mode_s_EXP
 /**
  * cogl_pipeline_get_layer_wrap_mode_s:
  * @pipeline: A #CoglPipeline object
@@ -380,13 +380,12 @@ cogl_pipeline_get_layer_point_sprite_coords_enabled (CoglPipeline *pipeline,
  * this layer.
  *
  * Since: 1.6
+ * Stability: unstable
  */
 CoglPipelineWrapMode
 cogl_pipeline_get_layer_wrap_mode_s (CoglPipeline *pipeline,
                                      int           layer_index);
 
-#define cogl_pipeline_set_layer_wrap_mode_s \
-  cogl_pipeline_set_layer_wrap_mode_s_EXP
 /**
  * cogl_pipeline_set_layer_wrap_mode_s:
  * @pipeline: A #CoglPipeline object
@@ -396,14 +395,13 @@ cogl_pipeline_get_layer_wrap_mode_s (CoglPipeline *pipeline,
  * Sets the wrap mode for the 's' coordinate of texture lookups on this layer.
  *
  * Since: 2.0
+ * Stability: unstable
  */
 void
 cogl_pipeline_set_layer_wrap_mode_s (CoglPipeline        *pipeline,
                                      int                  layer_index,
                                      CoglPipelineWrapMode mode);
 
-#define cogl_pipeline_get_layer_wrap_mode_t \
-  cogl_pipeline_get_layer_wrap_mode_t_EXP
 /**
  * cogl_pipeline_get_layer_wrap_mode_t:
  * @pipeline: A #CoglPipeline object
@@ -416,14 +414,13 @@ cogl_pipeline_set_layer_wrap_mode_s (CoglPipeline        *pipeline,
  * this layer.
  *
  * Since: 1.6
+ * Stability: unstable
  */
 CoglPipelineWrapMode
 cogl_pipeline_get_layer_wrap_mode_t (CoglPipeline *pipeline,
                                      int           layer_index);
 
 
-#define cogl_pipeline_set_layer_wrap_mode_t \
-  cogl_pipeline_set_layer_wrap_mode_t_EXP
 /**
  * cogl_pipeline_set_layer_wrap_mode_t:
  * @pipeline: A #CoglPipeline object
@@ -433,14 +430,13 @@ cogl_pipeline_get_layer_wrap_mode_t (CoglPipeline *pipeline,
  * Sets the wrap mode for the 't' coordinate of texture lookups on this layer.
  *
  * Since: 2.0
+ * Stability: unstable
  */
 void
 cogl_pipeline_set_layer_wrap_mode_t (CoglPipeline        *pipeline,
                                      int                  layer_index,
                                      CoglPipelineWrapMode mode);
 
-#define cogl_pipeline_get_layer_wrap_mode_p \
-  cogl_pipeline_get_layer_wrap_mode_p_EXP
 /**
  * cogl_pipeline_get_layer_wrap_mode_p:
  * @pipeline: A #CoglPipeline object
@@ -453,13 +449,12 @@ cogl_pipeline_set_layer_wrap_mode_t (CoglPipeline        *pipeline,
  * this layer.
  *
  * Since: 1.6
+ * Stability: unstable
  */
 CoglPipelineWrapMode
 cogl_pipeline_get_layer_wrap_mode_p (CoglPipeline *pipeline,
                                      int           layer_index);
 
-#define cogl_pipeline_set_layer_wrap_mode_p \
-  cogl_pipeline_set_layer_wrap_mode_p_EXP
 /**
  * cogl_pipeline_set_layer_wrap_mode_p:
  * @pipeline: A #CoglPipeline object
@@ -470,14 +465,13 @@ cogl_pipeline_get_layer_wrap_mode_p (CoglPipeline *pipeline,
  * this layer. 'p' is the third coordinate.
  *
  * Since: 2.0
+ * Stability: unstable
  */
 void
 cogl_pipeline_set_layer_wrap_mode_p (CoglPipeline        *pipeline,
                                      int                  layer_index,
                                      CoglPipelineWrapMode mode);
 
-#define cogl_pipeline_set_layer_wrap_mode \
-  cogl_pipeline_set_layer_wrap_mode_EXP
 /**
  * cogl_pipeline_set_layer_wrap_mode:
  * @pipeline: A #CoglPipeline object
@@ -491,12 +485,33 @@ cogl_pipeline_set_layer_wrap_mode_p (CoglPipeline        *pipeline,
  * cogl_pipeline_set_layer_wrap_mode_p() separately.
  *
  * Since: 2.0
+ * Stability: unstable
  */
 void
 cogl_pipeline_set_layer_wrap_mode (CoglPipeline        *pipeline,
                                    int                  layer_index,
                                    CoglPipelineWrapMode mode);
 
+/**
+ * cogl_pipeline_add_layer_snippet:
+ * @pipeline: A #CoglPipeline
+ * @layer: The layer to hook the snippet to
+ * @snippet: A #CoglSnippet
+ *
+ * Adds a shader snippet that will hook on to the given layer of the
+ * pipeline. The exact part of the pipeline that the snippet wraps
+ * around depends on the hook that is given to
+ * cogl_snippet_new(). Note that some hooks can't be used with a layer
+ * and need to be added with cogl_pipeline_add_snippet() instead.
+ *
+ * Since: 1.10
+ * Stability: Unstable
+ */
+void
+cogl_pipeline_add_layer_snippet (CoglPipeline *pipeline,
+                                 int layer,
+                                 CoglSnippet *snippet);
+
 #endif /* COGL_ENABLE_EXPERIMENTAL_API */
 
 G_END_DECLS
