@@ -43,9 +43,27 @@
 #include "cogl-display-private.h"
 #include "cogl-winsys-private.h"
 #include "cogl-winsys-stub-private.h"
-#include "cogl-winsys-egl-private.h"
 #include "cogl-config-private.h"
 
+#ifdef COGL_HAS_EGL_PLATFORM_XLIB_SUPPORT
+#include "cogl-winsys-egl-x11-private.h"
+#endif
+#ifdef COGL_HAS_EGL_PLATFORM_WAYLAND_SUPPORT
+#include "cogl-winsys-egl-wayland-private.h"
+#endif
+#ifdef COGL_HAS_EGL_PLATFORM_KMS_SUPPORT
+#include "cogl-winsys-egl-kms-private.h"
+#endif
+#ifdef COGL_HAS_EGL_PLATFORM_GDL_SUPPORT
+#include "cogl-winsys-egl-gdl-private.h"
+#endif
+#ifdef COGL_HAS_EGL_PLATFORM_ANDROID_SUPPORT
+#include "cogl-winsys-egl-android-private.h"
+#endif
+#ifdef COGL_HAS_EGL_PLATFORM_POWERVR_NULL_SUPPORT
+#include "cogl-winsys-egl-null-private.h"
+#endif
+
 #if COGL_HAS_XLIB_SUPPORT
 #include "cogl-xlib-renderer.h"
 #endif
@@ -56,6 +74,9 @@ extern const CoglWinsysVtable *_cogl_winsys_glx_get_vtable (void);
 #ifdef COGL_HAS_WGL_SUPPORT
 extern const CoglWinsysVtable *_cogl_winsys_wgl_get_vtable (void);
 #endif
+#ifdef COGL_HAS_SDL_SUPPORT
+extern const CoglWinsysVtable *_cogl_winsys_sdl_get_vtable (void);
+#endif
 
 typedef const CoglWinsysVtable *(*CoglWinsysVtableGetter) (void);
 
@@ -64,11 +85,29 @@ static CoglWinsysVtableGetter _cogl_winsys_vtable_getters[] =
 #ifdef COGL_HAS_GLX_SUPPORT
   _cogl_winsys_glx_get_vtable,
 #endif
-#ifdef COGL_HAS_EGL_SUPPORT
-  _cogl_winsys_egl_get_vtable,
+#ifdef COGL_HAS_EGL_PLATFORM_XLIB_SUPPORT
+  _cogl_winsys_egl_xlib_get_vtable,
+#endif
+#ifdef COGL_HAS_EGL_PLATFORM_WAYLAND_SUPPORT
+  _cogl_winsys_egl_wayland_get_vtable,
+#endif
+#ifdef COGL_HAS_EGL_PLATFORM_KMS_SUPPORT
+  _cogl_winsys_egl_kms_get_vtable,
+#endif
+#ifdef COGL_HAS_EGL_PLATFORM_GDL_SUPPORT
+  _cogl_winsys_egl_gdl_get_vtable,
+#endif
+#ifdef COGL_HAS_EGL_PLATFORM_ANDROID_SUPPORT
+  _cogl_winsys_egl_android_get_vtable,
+#endif
+#ifdef COGL_HAS_EGL_PLATFORM_POWERVR_NULL_SUPPORT
+  _cogl_winsys_egl_null_get_vtable,
 #endif
 #ifdef COGL_HAS_WGL_SUPPORT
   _cogl_winsys_wgl_get_vtable,
+#endif
+#ifdef COGL_HAS_SDL_SUPPORT
+  _cogl_winsys_sdl_get_vtable,
 #endif
   _cogl_winsys_stub_get_vtable,
 };
@@ -130,6 +169,10 @@ cogl_renderer_new (void)
   renderer->connected = FALSE;
   renderer->event_filters = NULL;
 
+#ifdef COGL_HAS_XLIB_SUPPORT
+  renderer->xlib_enable_event_retrieval = TRUE;
+#endif
+
   return _cogl_renderer_object_new (renderer);
 }
 
@@ -144,6 +187,10 @@ cogl_xlib_renderer_set_foreign_display (CoglRenderer *renderer,
   _COGL_RETURN_IF_FAIL (!renderer->connected);
 
   renderer->foreign_xdpy = xdisplay;
+
+  /* If the application is using a foreign display then we can assume
+     it will also do its own event retrieval */
+  cogl_xlib_renderer_set_event_retrieval_enabled (renderer, FALSE);
 }
 
 Display *
@@ -153,6 +200,17 @@ cogl_xlib_renderer_get_foreign_display (CoglRenderer *renderer)
 
   return renderer->foreign_xdpy;
 }
+
+void
+cogl_xlib_renderer_set_event_retrieval_enabled (CoglRenderer *renderer,
+                                                gboolean enable)
+{
+  _COGL_RETURN_IF_FAIL (cogl_is_renderer (renderer));
+  /* NB: Renderers are considered immutable once connected */
+  _COGL_RETURN_IF_FAIL (!renderer->connected);
+
+  renderer->xlib_enable_event_retrieval = enable;
+}
 #endif /* COGL_HAS_XLIB_SUPPORT */
 
 gboolean
@@ -161,9 +219,8 @@ cogl_renderer_check_onscreen_template (CoglRenderer *renderer,
                                        GError **error)
 {
   CoglDisplay *display;
-  const CoglWinsysVtable *winsys = _cogl_renderer_get_winsys (renderer);
 
-  if (!winsys->renderer_connect (renderer, error))
+  if (!cogl_renderer_connect (renderer, error))
     return FALSE;
 
   display = cogl_display_new (renderer, onscreen_template);
@@ -189,7 +246,7 @@ _cogl_renderer_choose_driver (CoglRenderer *renderer,
     driver_name = _cogl_config_driver;
 
 #ifdef HAVE_COGL_GL
-  if (driver_name == NULL || !strcmp (driver_name, "gl"))
+  if (driver_name == NULL || !g_ascii_strcasecmp (driver_name, "gl"))
     {
       renderer->driver = COGL_DRIVER_GL;
       libgl_name = COGL_GL_LIBNAME;
@@ -198,7 +255,7 @@ _cogl_renderer_choose_driver (CoglRenderer *renderer,
 #endif
 
 #ifdef HAVE_COGL_GLES2
-  if (driver_name == NULL || !strcmp (driver_name, "gles2"))
+  if (driver_name == NULL || !g_ascii_strcasecmp (driver_name, "gles2"))
     {
       renderer->driver = COGL_DRIVER_GLES2;
       libgl_name = COGL_GLES2_LIBNAME;
@@ -207,7 +264,7 @@ _cogl_renderer_choose_driver (CoglRenderer *renderer,
 #endif
 
 #ifdef HAVE_COGL_GLES
-  if (driver_name == NULL || !strcmp (driver_name, "gles1"))
+  if (driver_name == NULL || !g_ascii_strcasecmp (driver_name, "gles1"))
     {
       renderer->driver = COGL_DRIVER_GLES1;
       libgl_name = COGL_GLES1_LIBNAME;
@@ -264,6 +321,8 @@ cogl_renderer_connect (CoglRenderer *renderer, GError **error)
     {
       const CoglWinsysVtable *winsys = _cogl_winsys_vtable_getters[i]();
       GError *tmp_error = NULL;
+      GList *l;
+      gboolean constraints_failed = FALSE;
 
       if (renderer->winsys_id_override != COGL_WINSYS_ID_ANY)
         {
@@ -275,10 +334,23 @@ cogl_renderer_connect (CoglRenderer *renderer, GError **error)
           char *user_choice = getenv ("COGL_RENDERER");
           if (!user_choice)
             user_choice = _cogl_config_renderer;
-          if (user_choice && strcmp (winsys->name, user_choice) != 0)
+          if (user_choice &&
+              g_ascii_strcasecmp (winsys->name, user_choice) != 0)
             continue;
         }
 
+      for (l = renderer->constraints; l; l = l->next)
+        {
+          CoglRendererConstraint constraint = GPOINTER_TO_UINT (l->data);
+          if (!(winsys->constraints & constraint))
+            {
+              constraints_failed = TRUE;
+              break;
+            }
+        }
+      if (constraints_failed)
+        continue;
+
       /* At least temporarily we will associate this winsys with
        * the renderer in-case ->renderer_connect calls API that
        * wants to query the current winsys... */
@@ -415,3 +487,22 @@ cogl_renderer_get_n_fragment_texture_units (CoglRenderer *renderer)
 
   return n;
 }
+
+void
+cogl_renderer_add_contraint (CoglRenderer *renderer,
+                             CoglRendererConstraint constraint)
+{
+  g_return_if_fail (!renderer->connected);
+  renderer->constraints = g_list_prepend (renderer->constraints,
+                                          GUINT_TO_POINTER (constraint));
+}
+
+void
+cogl_renderer_remove_constraint (CoglRenderer *renderer,
+                                 CoglRendererConstraint constraint)
+{
+  g_return_if_fail (!renderer->connected);
+  renderer->constraints = g_list_remove (renderer->constraints,
+                                         GUINT_TO_POINTER (constraint));
+}
+
