@@ -30,6 +30,10 @@
 
 #include <cogl/cogl-types.h>
 #include <cogl/cogl-defines.h>
+#if defined (COGL_ENABLE_EXPERIMENTAL_API)
+#include <cogl/cogl-pixel-buffer.h>
+#endif
+#include <cogl/cogl-bitmap.h>
 
 G_BEGIN_DECLS
 
@@ -37,11 +41,14 @@ G_BEGIN_DECLS
  * SECTION:cogl-texture
  * @short_description: Fuctions for creating and manipulating textures
  *
- * COGL allows creating and manipulating GL textures using a uniform
+ * Cogl allows creating and manipulating textures using a uniform
  * API that tries to hide all the various complexities of creating,
  * loading and manipulating textures.
  */
 
+typedef struct _CoglTexture CoglTexture;
+#define COGL_TEXTURE(X) ((CoglTexture *)X)
+
 #define COGL_TEXTURE_MAX_WASTE  127
 
 /**
@@ -67,7 +74,8 @@ G_BEGIN_DECLS
 typedef enum {
   COGL_TEXTURE_ERROR_SIZE,
   COGL_TEXTURE_ERROR_FORMAT,
-  COGL_TEXTURE_ERROR_BAD_PARAMETER
+  COGL_TEXTURE_ERROR_BAD_PARAMETER,
+  COGL_TEXTURE_ERROR_TYPE
 } CoglTextureError;
 
 GQuark cogl_texture_error_quark (void);
@@ -80,14 +88,13 @@ GQuark cogl_texture_error_quark (void);
  * @internal_format: the #CoglPixelFormat to use for the GPU storage of the
  *    texture.
  *
- * Creates a new COGL texture with the specified dimensions and pixel format.
+ * Creates a new #CoglTexture with the specified dimensions and pixel format.
  *
- * Return value: a #CoglHandle to the newly created texture or
- *   %COGL_INVALID_HANDLE on failure
+ * Return value: A newly created #CoglTexture or %NULL on failure
  *
  * Since: 0.8
  */
-CoglHandle
+CoglTexture *
 cogl_texture_new_with_size (unsigned int width,
                             unsigned int height,
                             CoglTextureFlags flags,
@@ -107,14 +114,13 @@ cogl_texture_new_with_size (unsigned int width,
  *    other than straight blending.
  * @error: return location for a #GError or %NULL
  *
- * Creates a COGL texture from an image file.
+ * Creates a #CoglTexture from an image file.
  *
- * Return value: a #CoglHandle to the newly created texture or
- *    %COGL_INVALID_HANDLE on failure
+ * Return value: A newly created #CoglTexture or %NULL on failure
  *
  * Since: 0.8
  */
-CoglHandle
+CoglTexture *
 cogl_texture_new_from_file (const char       *filename,
                             CoglTextureFlags   flags,
                             CoglPixelFormat    internal_format,
@@ -138,14 +144,13 @@ cogl_texture_new_from_file (const char       *filename,
  *    scanlines in @data
  * @data: pointer the memory region where the source buffer resides
  *
- * Creates a new COGL texture based on data residing in memory.
+ * Creates a new #CoglTexture based on data residing in memory.
  *
- * Return value: a #CoglHandle to the newly created texture or
- *   %COGL_INVALID_HANDLE on failure
+ * Return value: A newly created #CoglTexture or %NULL on failure
  *
  * Since: 0.8
  */
-CoglHandle
+CoglTexture *
 cogl_texture_new_from_data (unsigned int      width,
                             unsigned int      height,
                             CoglTextureFlags  flags,
@@ -164,7 +169,7 @@ cogl_texture_new_from_data (unsigned int      width,
  * @y_pot_waste: vertical waste on the bottom edge of the texture.
  * @format: format of the foreign texture.
  *
- * Creates a COGL texture based on an existing OpenGL texture; the
+ * Creates a #CoglTexture based on an existing OpenGL texture; the
  * width, height and format are passed along since it is not always
  * possible to query these from OpenGL.
  *
@@ -175,12 +180,11 @@ cogl_texture_new_from_data (unsigned int      width,
  * the waste arguments to tell Cogl which region should be mapped to
  * the texture coordinate range [0:1].
  *
- * Return value: a #CoglHandle to the newly created texture or
- *   %COGL_INVALID_HANDLE on failure
+ * Return value: A newly created #CoglTexture or %NULL on failure
  *
  * Since: 0.8
  */
-CoglHandle
+CoglTexture *
 cogl_texture_new_from_foreign (GLuint          gl_handle,
                                GLenum          gl_target,
                                GLuint          width,
@@ -191,83 +195,102 @@ cogl_texture_new_from_foreign (GLuint          gl_handle,
 
 /**
  * cogl_texture_new_from_bitmap:
- * @bmp_handle: A CoglBitmap handle
+ * @bitmap: A #CoglBitmap pointer
  * @flags: Optional flags for the texture, or %COGL_TEXTURE_NONE
  * @internal_format: the #CoglPixelFormat to use for the GPU storage of the
  * texture
  *
- * Creates a COGL texture from a CoglBitmap.
+ * Creates a #CoglTexture from a #CoglBitmap.
  *
- * Return value: a #CoglHandle to the newly created texture or
- *   %COGL_INVALID_HANDLE on failure
+ * Return value: A newly created #CoglTexture or %NULL on failure
  *
  * Since: 1.0
  */
-CoglHandle
-cogl_texture_new_from_bitmap (CoglHandle       bmp_handle,
+CoglTexture *
+cogl_texture_new_from_bitmap (CoglBitmap *bitmap,
                               CoglTextureFlags flags,
-                              CoglPixelFormat  internal_format);
+                              CoglPixelFormat internal_format);
 
 /**
  * cogl_is_texture:
- * @handle: A CoglHandle
+ * @object: A #CoglObject pointer
  *
- * Gets whether the given handle references an existing texture object.
+ * Gets whether the given object references a texture object.
  *
  * Return value: %TRUE if the handle references a texture, and
  *   %FALSE otherwise
  */
 gboolean
-cogl_is_texture (CoglHandle handle);
+cogl_is_texture (void *object);
 
 /**
  * cogl_texture_get_width:
- * @handle: a #CoglHandle for a texture.
+ * @texture a #CoglTexture pointer.
  *
  * Queries the width of a cogl texture.
  *
  * Return value: the width of the GPU side texture in pixels
  */
 unsigned int
-cogl_texture_get_width (CoglHandle handle);
+cogl_texture_get_width (CoglTexture *texture);
 
 /**
  * cogl_texture_get_height:
- * @handle: a #CoglHandle for a texture.
+ * @texture a #CoglTexture pointer.
  *
  * Queries the height of a cogl texture.
  *
  * Return value: the height of the GPU side texture in pixels
  */
 unsigned int
-cogl_texture_get_height (CoglHandle handle);
+cogl_texture_get_height (CoglTexture *texture);
 
 /**
  * cogl_texture_get_format:
- * @handle: a #CoglHandle for a texture.
+ * @texture a #CoglTexture pointer.
  *
  * Queries the #CoglPixelFormat of a cogl texture.
  *
  * Return value: the #CoglPixelFormat of the GPU side texture
  */
 CoglPixelFormat
-cogl_texture_get_format (CoglHandle handle);
+cogl_texture_get_format (CoglTexture *texture);
 
 
 /**
  * cogl_texture_get_rowstride:
- * @handle: a #CoglHandle for a texture.
+ * @texture a #CoglTexture pointer.
  *
- * Queries the rowstride of a cogl texture.
+ * Determines the bytes-per-pixel for the #CoglPixelFormat retrieved
+ * from cogl_texture_get_format() and multiplies that by the texture's
+ * width.
  *
- * Return value: the offset in bytes between each consequetive row of pixels
+ * <note>It's very unlikely that anyone would need to use this API to
+ * query the internal rowstride of a #CoglTexture which can just be
+ * considered an implementation detail. Actually it's not even useful
+ * internally since underlying drivers are free to use a different
+ * format</note>
+ *
+ * </note>This API is only here for backwards compatibility and
+ * shouldn't be used in new code. In particular please don't be
+ * mislead to pass the returned value to cogl_texture_get_data() for
+ * the rowstride, since you should be passing the rowstride you desire
+ * for your destination buffer not the rowstride of the source
+ * texture.</note>
+ *
+ * Return value: The bytes-per-pixel for the current format
+ *               multiplied by the texture's width
+ *
+ * Deprecated: 1.10: There's no replacement for the API but there's
+ *                   also no known need for API either. It was just
+ *                   a mistake that it was ever published.
  */
 unsigned int
-cogl_texture_get_rowstride (CoglHandle handle);
+cogl_texture_get_rowstride (CoglTexture *texture);
 
 /**
  * cogl_texture_get_max_waste:
- * @handle: a #CoglHandle for a texture.
+ * @texture a #CoglTexture pointer.
  *
  * Queries the maximum wasted (unused) pixels in one dimension of a GPU side
  * texture.
@@ -275,11 +298,11 @@ cogl_texture_get_rowstride (CoglHandle handle);
  * Return value: the maximum waste
  */
 int
-cogl_texture_get_max_waste (CoglHandle handle);
+cogl_texture_get_max_waste (CoglTexture *texture);
 
 /**
  * cogl_texture_is_sliced:
- * @handle: a #CoglHandle for a texture.
+ * @texture a #CoglTexture pointer.
  *
  * Queries if a texture is sliced (stored as multiple GPU side tecture
  * objects).
@@ -288,17 +311,17 @@ cogl_texture_get_max_waste (CoglHandle handle);
  *   is stored as a single GPU texture
  */
 gboolean
-cogl_texture_is_sliced (CoglHandle handle);
+cogl_texture_is_sliced (CoglTexture *texture);
 
 /**
  * cogl_texture_get_gl_texture:
- * @handle: a #CoglHandle for a texture.
+ * @texture a #CoglTexture pointer.
  * @out_gl_handle: (out) (allow-none): pointer to return location for the
  *   textures GL handle, or %NULL.
  * @out_gl_target: (out) (allow-none): pointer to return location for the
  *   GL target type, or %NULL.
  *
- * Queries the GL handles for a GPU side texture through its #CoglHandle.
+ * Queries the GL handles for a GPU side texture through its #CoglTexture.
  *
  * If the texture is spliced the data for the first sub texture will be
  * queried.
@@ -307,39 +330,46 @@ cogl_texture_is_sliced (CoglHandle handle);
  *   if the handle was invalid
  */
 gboolean
-cogl_texture_get_gl_texture (CoglHandle   handle,
+cogl_texture_get_gl_texture (CoglTexture *texture,
                              GLuint      *out_gl_handle,
                              GLenum      *out_gl_target);
 
 /**
  * cogl_texture_get_data:
- * @handle: a #CoglHandle for a texture.
+ * @texture a #CoglTexture pointer.
  * @format: the #CoglPixelFormat to store the texture as.
- * @rowstride: the rowstride of @data or retrieved from texture if none is
- * specified.
- * @data: memory location to write contents of buffer, or %NULL if we're
- * only querying the data size through the return value.
+ * @rowstride: the rowstride of @data in bytes or pass 0 to calculate
+ *             from the bytes-per-pixel of @format multiplied by the
+ *             @texture width.
+ * @data: memory location to write the @texture's contents, or %NULL
+ * to only query the data size through the return value.
  *
  * Copies the pixel data from a cogl texture to system memory.
  *
- * Return value: the size of the texture data in bytes, or 0 if the texture
- *   is not valid
+ * <note>Don't pass the value of cogl_texture_get_rowstride() as the
+ * @rowstride argument, the rowstride should be the rowstride you
+ * want for the destination @data buffer not the rowstride of the
+ * source texture</note>
+ *
+ * Return value: the size of the texture data in bytes
  */
 int
-cogl_texture_get_data (CoglHandle       handle,
-                       CoglPixelFormat  format,
-                       unsigned int     rowstride,
-                       guint8          *data);
+cogl_texture_get_data (CoglTexture *texture,
+                       CoglPixelFormat format,
+                       unsigned int rowstride,
+                       guint8 *data);
 
 /**
  * cogl_texture_set_region:
- * @handle: a #CoglHandle.
+ * @texture a #CoglTexture.
  * @src_x: upper left coordinate to use from source data.
  * @src_y: upper left coordinate to use from source data.
  * @dst_x: upper left destination horizontal coordinate.
  * @dst_y: upper left destination vertical coordinate.
- * @dst_width: width of destination region to write.
- * @dst_height: height of destination region to write.
+ * @dst_width: width of destination region to write. (Must be less
+ *   than or equal to @width)
+ * @dst_height: height of destination region to write. (Must be less
+ *   than or equal to @height)
  * @width: width of source data buffer.
  * @height: height of source data buffer.
  * @format: the #CoglPixelFormat used in the source buffer.
@@ -347,14 +377,16 @@ cogl_texture_get_data (CoglHandle       handle,
  * specified)
  * @data: the actual pixel data.
  *
- * Sets the pixels in a rectangular subregion of @handle from an in-memory
+ * Sets the pixels in a rectangular subregion of @texture from an in-memory
  * buffer containing pixel data.
  *
+ * <note>The region set can't be larger than the source @data</note>
+ *
  * Return value: %TRUE if the subregion upload was successful, and
  *   %FALSE otherwise
  */
 gboolean
-cogl_texture_set_region (CoglHandle       handle,
+cogl_texture_set_region (CoglTexture     *texture,
                          int              src_x,
                          int              src_y,
                          int              dst_x,
@@ -367,9 +399,49 @@ cogl_texture_set_region (CoglHandle       handle,
                          unsigned int     rowstride,
                          const guint8    *data);
 
+#if defined (COGL_ENABLE_EXPERIMENTAL_API)
+
+#define cogl_texture_set_region_from_bitmap \
+  cogl_texture_set_region_from_bitmap_EXP
+/**
+ * cogl_texture_set_region_from_bitmap:
+ * @texture a #CoglTexture pointer
+ * @src_x: upper left coordinate to use from the source bitmap.
+ * @src_y: upper left coordinate to use from the source bitmap
+ * @dst_x: upper left destination horizontal coordinate.
+ * @dst_y: upper left destination vertical coordinate.
+ * @dst_width: width of destination region to write. (Must be less
+ *   than or equal to the bitmap width)
+ * @dst_height: height of destination region to write. (Must be less
+ *   than or equal to the bitmap height)
+ * @bitmap: The source bitmap to read from
+ *
+ * Copies a specified source region from @bitmap to the position
+ * (@src_x, @src_y) of the given destination texture @handle.
+ *
+ * <note>The region updated can't be larger than the source
+ * bitmap</note>
+ *
+ * Return value: %TRUE if the subregion upload was successful, and
+ *   %FALSE otherwise
+ *
+ * Since: 1.8
+ * Stability: unstable
+ */
+gboolean
+cogl_texture_set_region_from_bitmap (CoglTexture *texture,
+                                     int src_x,
+                                     int src_y,
+                                     int dst_x,
+                                     int dst_y,
+                                     unsigned int dst_width,
+                                     unsigned int dst_height,
+                                     CoglBitmap *bitmap);
+#endif
+
 /**
  * cogl_texture_new_from_sub_texture:
- * @full_texture: a #CoglHandle to an existing texture
+ * @full_texture: a #CoglTexture pointer
  * @sub_x: X coordinate of the top-left of the subregion
  * @sub_y: Y coordinate of the top-left of the subregion
  * @sub_width: Width in pixels of the subregion
@@ -387,22 +459,22 @@ cogl_texture_set_region (CoglHandle       handle,
  * not need to keep one separately if you only want to use the sub
  * texture.
  *
- * Return value: a #CoglHandle to the new texture.
- *
+ * Return value: A newly created #CoglTexture or %NULL on failure
  * Since: 1.2
  */
-CoglHandle
-cogl_texture_new_from_sub_texture (CoglHandle full_texture,
-                                   int        sub_x,
-                                   int        sub_y,
-                                   int        sub_width,
-                                   int        sub_height);
+CoglTexture *
+cogl_texture_new_from_sub_texture (CoglTexture *full_texture,
+                                   int sub_x,
+                                   int sub_y,
+                                   int sub_width,
+                                   int sub_height);
 
 #if defined (COGL_ENABLE_EXPERIMENTAL_API)
 
+#define cogl_texture_new_from_buffer cogl_texture_new_from_buffer_EXP
 /**
  * cogl_texture_new_from_buffer:
- * @buffer: the #CoglHandle of a pixel buffer
+ * @buffer: A #CoglPixelBuffer pointer
  * @width: width of texture in pixels or 0
  * @height: height of texture in pixels or 0
  * @flags: optional flags for the texture, or %COGL_TEXTURE_NONE
@@ -424,14 +496,13 @@ cogl_texture_new_from_sub_texture (CoglHandle full_texture,
  * has been created using cogl_pixel_buffer_new_for_size() it's possible to omit
  * the height and width values already specified at creation time.
  *
- * Return value: a #CoglHandle to the new texture or %COGL_INVALID_HANDLE on
- *    failure
+ * Return value: A newly created #CoglTexture or %NULL on failure
  *
  * Since: 1.2
  * Stability: Unstable
  */
-CoglHandle
-cogl_texture_new_from_buffer  (CoglHandle       buffer,
+CoglTexture *
+cogl_texture_new_from_buffer  (CoglPixelBuffer *buffer,
                                unsigned int     width,
                                unsigned int     height,
                                CoglTextureFlags flags,
@@ -439,51 +510,33 @@ cogl_texture_new_from_buffer  (CoglHandle       buffer,
                                CoglPixelFormat  internal_format,
                                unsigned int     rowstride,
                                unsigned int     offset);
-
-/* the function above is experimental, the actual symbol is suffixed by _EXP so
- * we can ensure ABI compatibility and leave the cogl_buffer namespace free for
- * future use. A bunch of defines translates the symbols documented above into
- * the real symbols */
-
-CoglHandle
-cogl_texture_new_from_buffer_EXP (CoglHandle       buffer,
-                                  unsigned int     width,
-                                  unsigned int     height,
-                                  CoglTextureFlags flags,
-                                  CoglPixelFormat  format,
-                                  CoglPixelFormat  internal_format,
-                                  unsigned int     rowstride,
-                                  unsigned int     offset);
-
-#define cogl_texture_new_from_buffer cogl_texture_new_from_buffer_EXP
-
 #endif
 
 #ifndef COGL_DISABLE_DEPRECATED
 
 /**
  * cogl_texture_ref:
- * @handle: a @CoglHandle.
+ * @texture: a #CoglTexture.
  *
  * Increment the reference count for a cogl texture.
  *
- * Deprecated: 1.2: Use cogl_handle_ref() instead
+ * Deprecated: 1.2: Use cogl_object_ref() instead
  *
- * Return value: the @handle.
+ * Return value: the @texture pointer.
  */
-CoglHandle
-cogl_texture_ref (CoglHandle handle) G_GNUC_DEPRECATED;
+void *
+cogl_texture_ref (void *texture) G_GNUC_DEPRECATED;
 
 /**
  * cogl_texture_unref:
- * @handle: a @CoglHandle.
+ * @texture: a #CoglTexture.
  *
  * Decrement the reference count for a cogl texture.
  *
- * Deprecated: 1.2: Use cogl_handle_unref() instead
+ * Deprecated: 1.2: Use cogl_object_unref() instead
  */
 void
-cogl_texture_unref (CoglHandle handle) G_GNUC_DEPRECATED;
+cogl_texture_unref (void *texture) G_GNUC_DEPRECATED;
 
 #endif /* COGL_DISABLE_DEPRECATED */
 
