@@ -88,6 +88,10 @@ main (int argc, char **argv)
     pipeline = cogl_pipeline_new ();
 
     for (;;) {
+        CoglPollFD *poll_fds;
+        int n_poll_fds;
+        gint64 timeout;
+
         cogl_clear (&black, COGL_BUFFER_BIT_COLOR);
 
         cogl_push_matrix ();
@@ -106,6 +110,10 @@ main (int argc, char **argv)
         cogl_rectangle (0, 1, 1, -1);
 
         cogl_framebuffer_swap_buffers (fb);
+
+        cogl_poll_get_info (ctx, &poll_fds, &n_poll_fds, &timeout);
+        g_poll ((GPollFD *) poll_fds, n_poll_fds, 0);
+        cogl_poll_dispatch (ctx, poll_fds, n_poll_fds);
     }
 
     return 0;
