@@ -15,26 +15,6 @@ typedef struct _TestState
   CoglFramebuffer *fbo[NUM_FBOS];
 } TestState;
 
-static void
-check_pixel (int x, int y, guint8 r, guint8 g, guint8 b)
-{
-  guint32 pixel;
-  char *screen_pixel;
-  char *intended_pixel;
-
-  cogl_read_pixels (x, y, 1, 1, COGL_READ_PIXELS_COLOR_BUFFER,
-                    COGL_PIXEL_FORMAT_RGBA_8888_PRE,
-                    (guint8 *) &pixel);
-
-  screen_pixel = g_strdup_printf ("#%06x", GUINT32_FROM_BE (pixel) >> 8);
-  intended_pixel = g_strdup_printf ("#%02x%02x%02x", r, g, b);
-
-  g_assert_cmpstr (screen_pixel, ==, intended_pixel);
-
-  g_free (screen_pixel);
-  g_free (intended_pixel);
-}
-
 static void
 paint (TestState *state)
 {
@@ -78,11 +58,11 @@ paint (TestState *state)
           { 0x00, 0xff, 0x00, 0xff },
           { 0x00, 0x00, 0xff, 0xff } };
 
-      check_pixel (state->width * (i + 0.5f) / NUM_FBOS,
-                   state->height / 2,
-                   expected_colors[i][0],
-                   expected_colors[i][1],
-                   expected_colors[i][2]);
+      test_utils_check_pixel_rgb (state->width * (i + 0.5f) / NUM_FBOS,
+                                  state->height / 2,
+                                  expected_colors[i][0],
+                                  expected_colors[i][1],
+                                  expected_colors[i][2]);
     }
 }
 
